import java.io.*;
import java.util.*;

/**
 * Предоставляет доступ к множеству маленьких файлов, находящихся в одной
 * директории, точно к одному большому файлу. Максимальный размер одного члена
 * колонии задан в {@link EasyReadingFilesColony#MAX_FILE_SIZE}. По мере
 * чтения открываются нужные файлы-члены колонии. Членами колонии должны быть
 * файлы с именами-шестнадцатиричными числами ("0", "1" ... "9", "a", "b" ...
 * "e", "f", "10" ... ).
 *
 * @author quasiyoke
 */
public class EasyReadingFilesColony extends EasyReadingFile{

private static final int MAX_FILE_SIZE=20*1024;

private Vector files=new Vector();

private String folderPath;

/**
 * Если папка отсутствует, {@link IOException} возникнет лишь при первой
 * попытке чтения.
 */
public EasyReadingFilesColony(String folderPath){
    this.folderPath=folderPath;
}

public void close() throws IOException{
    int j;
    int filesCount=files.size();
    for(j=0;j<filesCount;j++)
	((EasyReadingSolitaryFile)files.elementAt(j)).close();
}

private EasyReadingSolitaryFile getDataInput() throws IOException{
    EasyReadingSolitaryFile dataInput;
    int fileNumber=(int)(filePointer/MAX_FILE_SIZE);
    boolean resizingNeeds=fileNumber>=files.size();
    if(resizingNeeds||(dataInput=(EasyReadingSolitaryFile)files.elementAt(
	    fileNumber))==null){
	if(resizingNeeds)
	    files.setSize(fileNumber+1);
	files.setElementAt(dataInput=new EasyReadingSolitaryFile(folderPath+
		'/'+Integer.toHexString(fileNumber)),fileNumber);
    }
    dataInput.seek(filePointer%MAX_FILE_SIZE);
    return dataInput;
}

public char readChar() throws IOException{
    return (char)((readUnsignedByte()<<8)|readUnsignedByte());
}

public int readInt() throws IOException{
    return (readUnsignedByte()<<24)|(readUnsignedByte()<<16)|
	    (readUnsignedByte()<<8)|readUnsignedByte();
}

public int readUnsignedByte() throws IOException{
    int result=getDataInput().readUnsignedByte();
    filePointer++;
    return result;
}

public EasyReadingFile seek(long position){
    filePointer=position;
    return this;
}

public EasyReadingFile skipBytes(int bytesCount) throws IOException{
    if(bytesCount>0)
	filePointer+=bytesCount;
    return this;
}

}
