import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.rms.*;
import java.util.*;
import java.io.*;
import org.xmlpull.v1.*;

/**
 * Требующиеся для работы файлы:
 * <ul>
 * <li>
 * Список путей до папок-словарей в UTF-8 файле с путём {@link
 * LexicalBase#DICTIONARIES_LIST_FILE_PATH}.<br>
 * Формат файла ::= (Путь_до_папки_словаря '\n')+
 * </ul>
 * @author quasiyoke
 */
public final class PieceOfDictionary extends MIDlet implements CommandListener{

private final static int ALERT_TIMEOUT=4000;

public final static int BYTE_LENGTH=1;

public final static int CHAR_LENGTH=2;

public final static int INT_LENGTH=4;

public final static String MONEYBOXES=
        "Z182466285476, E362211144058, R156368550112";

public final static int POINT_POSITION=6;

private final static String PARAMETERS_FILE_PATH="parameters.xml";

private final static String RECORD_STORE_NAME="PieceOfDictionaryStorage";

private final static String RECORD_STORE_NOT_FOUND_MESSAGE=
        "RECORD_STORE_NOT_FOUND_MESSAGE";

/**
 * Отсортированный в <code>static{}</code> массив символов, считающихся
 * пробельными.
 */
private final static char[] SPACE_CHARACTERS={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,
    15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,' ','-','.',',',';','\'',
    '?','!','+',
};

/**
 * Строки, вместо которых следует автоматически подставлять некоторые строки в
 * готовые текстовые шаблоны.
 */
public final static String[] STRING_ESCAPE={"%s1","%s2","%s3",};

public final static int COLOR_BACKGROUND=0xFFFFFF;

private final static int COLOR_FOREGROUND=0;

public final static int COLOR_ARTICLES_KEYS=COLOR_FOREGROUND;

public final static int COLOR_ARTICLES_TEXT=COLOR_FOREGROUND;

public final static int COLOR_MESSAGES=middleColor(COLOR_FOREGROUND,
        COLOR_BACKGROUND,(1<<POINT_POSITION)/2);

public final static Font FONT_NORMAL=Font.getDefaultFont();

public final static int FONT_NORMAL_HEIGHT=FONT_NORMAL.getHeight();

public final static Font FONT_MESSAGES=FONT_NORMAL;

public final static Font FONT_ARTICLES_TEXT=FONT_NORMAL;

/**
 * Высота разделителя статей.
 */
public final static int ARTICLES_DIVIDER_HEIGHT=FONT_ARTICLES_TEXT.getHeight()*
        1/5;

public final static Font FONT_ARTICLES_KEYS=Font.getFont(FONT_NORMAL.getFace(),
        Font.STYLE_BOLD,FONT_NORMAL.getSize());

public final static Font FONT_ARTICLES_KEYS_HIGHLIGHTED=Font.getFont(
        FONT_NORMAL.getFace(),Font.STYLE_BOLD|Font.STYLE_UNDERLINED,FONT_NORMAL.
        getSize());

/**
 * Отсортированный (в <code>static{}</code>) массив соответствий меж кнопками
 * и буквами по умолчанию.
 */
private final static char[][] KEYS_CHARS_DEFAULT={
    {},/* 0 */
    {},/* 1 */
    {
        'a','b','c','а','б','в','г',
    },/* 2 */
    {
        'd','e','f','д','е','ё','ж','з',
    },/* 3 */
    {
        'g','h','i','и','й','к','л',
    },/* 4 */
    {
        'j','k','l','м','н','о','п',
    },/* 5 */
    {
        'm','n','o','р','с','т','у',
    },/* 6 */
    {
        'p','q','r','s','ф','х','ц','ч',
    },/* 7 */
    {
        't','u','v','ш','щ','ъ','ы',
    },/* 8 */
    {
        'w','x','y','z','ь','э','ю','я',
    },/* 9 */

};

private final static int KEYS_COUNT=KEYS_CHARS_DEFAULT.length;

public final static int MARGIN_LEFT=FONT_NORMAL_HEIGHT*1/7;

public final static int MARGIN_RIGHT=MARGIN_LEFT;

/**
 * Максимально число букв, закрепляемых за одной клавишей.
 */
private final static int MAX_KEY_CHARS_COUNT=25;

public final static int PI_FRACTION=(31416<<POINT_POSITION)/10000;

private static Display display;

private static PieceOfDictionary instance;

private char[][] keysChars;

private List languageSelect;

private String[] languagesNames;

private Command commandExit=new Command("Exit",Command.EXIT,100);

/**
 * Команда "Выбор языка". Должна встать в один ряд с инициализируемыми в {@link
 * PieceOfDictionary#initializeStrings()} кнопками.
 */
private Command commandSelectLanguage=new Command("Select language",
        Command.SCREEN,52);

private Command commandShowMainCanvas;

private Command commandOptionsSave;

private Command commandOptionsSetDefaults;

private Form formInformation;

private Form formOptions;

private ChoiceGroup formOptionsChoiceGroup;

private TextField[] formOptionsKeysChars;

private String menuItemExit;

private String menuItemInformation;

private String menuItemMinimize;

private String menuItemOptions;

private List menu;

private static final int VERSION_LENGTH=3;

/**
 * Версия программы.
 * <p>
 * {@code version[0]} -- Major<br>
 * {@code version[1]} -- Minor<br>
 * {@code version[2]} -- Revision
 */
private int version[]=new int[VERSION_LENGTH];

public static XmlPullParserFactory xmlPullParserFactory;



static{
    try{
        xmlPullParserFactory=XmlPullParserFactory.newInstance();
    }catch(XmlPullParserException e){
        showErrorAndExit(e.toString());
    }
    sort(SPACE_CHARACTERS);
    int j;
    for(j=0;j<KEYS_COUNT;j++){
        sort(KEYS_CHARS_DEFAULT[j]);
    }
}

public static int abs(int x){
    return x>0?x:-x;
}

public static Vector addAll(Vector source,int elementsCount,
        Vector destination){
    int j;
    for(j=0;j<elementsCount;j++){
        destination.addElement(source.elementAt(j));
    }
    return destination;
}

public void commandAction(Command command,Displayable displayable){
    int j;
    int i;
    if(command==commandShowMainCanvas){
        showMainCanvas();
    }else if(command==List.SELECT_COMMAND){
        String item;
        if(displayable==menu){
            item=menu.getString(menu.getSelectedIndex());
            if(item==menuItemInformation){
                display.setCurrent(formInformation);
            }else if(item==menuItemMinimize){
                minimize();
            }else if(item==menuItemExit){
                destroyApp(false);
            }else if(item==menuItemOptions){
                display.setCurrent(formOptions);
            }
        }else if(displayable==languageSelect){
            item=languageSelect.getString(languageSelect.getSelectedIndex());
            Language.setCurrent(item);
            deleteStorageAndSaveOptions();
            showMainCanvas();
        }
    }else if(command==commandOptionsSave){
        boolean[] selectedFlags=new boolean[1];
        formOptionsChoiceGroup.getSelectedFlags(selectedFlags);
        MainCanvas.getInstance().setFullScreenMode(MainCanvas.fullscreen=
                selectedFlags[0]);
        String sKeyChars;
        StringBuffer sbKeyChars;
        char character;
        int keyCharsCount;
        for(j=0;j<KEYS_COUNT;j++){
            sKeyChars=formOptionsKeysChars[j].getString();
            sbKeyChars=new StringBuffer();
            keyCharsCount=sKeyChars.length();
            for(i=0;i<keyCharsCount;i++){
                if(!isSpace(character=sKeyChars.charAt(i))){
                    /*
                     * TODO WTF?! toLowerCase не работает с кириллицей?! 
                     */
                    sbKeyChars.append(Character.toLowerCase(character));
                }
            }
            keysChars[j]=sort(sbKeyChars.toString().toCharArray());
        }
        deleteStorageAndSaveOptions();
        showMainCanvas();
        /*
         * Как можно больше делаем в фоне ПОСЛЕ показа экрана пользователю.
         */
        /*
         * Чтобы отобразились в опциях все sort'ы, trim'ы и toLowerCase'ы.
         */
        initializeOptionsFormItems();
    }else if(command==commandSelectLanguage){
        showLanguageSelect(true);
    }else if(command==commandOptionsSetDefaults){
        deleteStorageAndSaveDefaults();
    }else if(command==commandExit){
        destroyApp(false);
    }
}

public static int compareVersions(int[] version1,int[] version2){
    int j;
    for(j=0;j<VERSION_LENGTH;j++){
        if(version1[j]<version2[j]){
            return -1;
        }else if(version1[j]>version2[j]){
            return 1;
        }
    }
    return 0;
}

/**
 * Создаёт копию вектора.
 * @param vector Вектор-оригинал.
 * @param elementsCount Число элементов оригинала, подлежащих копированию.
 * @return Результат копирования.
 */
public static Vector copyVector(Vector vector,int elementsCount){
    return addAll(vector,elementsCount,new Vector(elementsCount));
}

private void deleteStorage(){
    try{
        RecordStore.deleteRecordStore(RECORD_STORE_NAME);
    }catch(RecordStoreException f){
    }
}

/**
 * Сохраняет в RMS опции по умолчанию. Делает запросы на пересопоставление
 * соответствий клавиш буквам и на переобработку последовательности клавиш.
 */
private void deleteStorageAndSaveDefaults(){
    int i;
    int j;
    keysChars=new char[KEYS_COUNT][];
    for(j=0;j<KEYS_COUNT;j++){
        keysChars[j]=new char[KEYS_CHARS_DEFAULT[j].length];
        for(i=0;i<KEYS_CHARS_DEFAULT[j].length;i++){
            keysChars[j][i]=KEYS_CHARS_DEFAULT[j][i];
        }
    }
    showLanguageSelect(false);
}

/**
 * Сохраняет в RMS опции. Делает запросы на пересопоставление соответствий
 * клавиш буквам и на переобработку последовательности клавиш.
 */
public void deleteStorageAndSaveOptions(){
    int j;
    int i;
    ByteArrayOutputStream byteArrayOutputStream=new ByteArrayOutputStream();
    DataOutputStream dataOutputStream=new DataOutputStream(
            byteArrayOutputStream);
    try{
        for(j=0;j<version.length;j++){
            dataOutputStream.writeInt(version[j]);
        }
        for(j=0;j<KEYS_COUNT;j++){
            dataOutputStream.writeByte(keysChars[j].length);
            for(i=0;i<keysChars[j].length;i++){
                dataOutputStream.writeChar(keysChars[j][i]);
            }
        }
        dataOutputStream.writeUTF(Language.getCurrentLanguageName());
        dataOutputStream.writeBoolean(MainCanvas.fullscreen);
        byte[] record=byteArrayOutputStream.toByteArray();
        deleteStorage();
        RecordStore recordStore=RecordStore.openRecordStore(RECORD_STORE_NAME,
                true);
        recordStore.addRecord(record,0,record.length);
        recordStore.closeRecordStore();
    }catch(IOException e){
        showError("Record preparing: "+e.toString());
    }catch(RecordStoreException e){
        showError("RMS writing: "+e.toString());
    }
    LexicalBase.getInstance().matchKeysCharsAndAlphabets(keysChars);
    MainCanvas.getInstance().processKeyChain();
}

public void destroyApp(boolean unconditional){
    display.setCurrent(null);
    System.out.println("Destroyed!");
    while(true){
        notifyDestroyed();
        try{
            Thread.sleep(100);
        }catch(InterruptedException e){
        }
    }
}

public static Display getDisplay(){
    return display;
}

public static PieceOfDictionary getInstance(){
    return instance;
}

/**
 * @return Строку вида "v. 1.0.23".
 */
public int[] getVersion(){
    return version;
}

/**
 * @return Строку вида "v. 1.0.23".
 */
public static String getVersion(int[] version){
    StringBuffer sbVersion=new StringBuffer("v. ");
    sbVersion.append(version[0]);
    sbVersion.append('.');
    sbVersion.append(version[1]);
    sbVersion.append('.');
    sbVersion.append(version[2]);
    return sbVersion.toString();
}

/**
 * @return Если {@code sVersion==null}, массив, проинициализированный нулями.
 */
public static int[] getVersion(String sVersion){
    int begin=0;
    int j;
    int i=0;
    int[] version=new int[VERSION_LENGTH];
    if(sVersion!=null){
        int sVersionLength=sVersion.length();
        for(j=0;j<VERSION_LENGTH;j++){
            for(;i<sVersionLength&&sVersion.charAt(i)!='.';i++){
            }
            if(i>begin){
                version[j]=Integer.parseInt(sVersion.substring(begin,i).trim());
                begin=++i;
            }else{
                break;
            }
        }
    }
    return version;
}

private void initializeOptionsFormItems(){
    int j;
    formOptions.deleteAll();
    String[] choiceGroupElements={Language.getCurrentLanguageInscription("fullscreen")};
    boolean[] choiceGroupElementsState={MainCanvas.fullscreen};
    formOptionsChoiceGroup=new ChoiceGroup("",ChoiceGroup.MULTIPLE,
            choiceGroupElements,null);
    formOptionsChoiceGroup.setSelectedFlags(choiceGroupElementsState);
    formOptions.append(formOptionsChoiceGroup);
    formOptions.append(Language.getCurrentLanguageInscription(
            "options_keys_help"));
    formOptionsKeysChars=new TextField[KEYS_COUNT];
    for(j=0;j<KEYS_COUNT;j++){
        formOptions.append(formOptionsKeysChars[j]=new TextField(Integer.
                toString(j),new String(keysChars[j]),MAX_KEY_CHARS_COUNT,TextField.SENSITIVE|
                TextField.NON_PREDICTIVE));
    }
}

/**
 * Инициализирует все зависимые от локализации поля.
 */
public void initializeStrings(){
    int j;
    commandShowMainCanvas=new Command(Language.getCurrentLanguageInscription(
            "show_main_screen"),Command.CANCEL,50);
    commandOptionsSave=new Command(Language.getCurrentLanguageInscription(
            "save_options"),Command.OK,51);
    /*
     * 52й порядок у commandSelectLanguage.
     */
    commandOptionsSetDefaults=
            new Command(Language.getCurrentLanguageInscription("set_defaults"),
            Command.OK,53);

    menuItemInformation=Language.getCurrentLanguageInscription(
            "menu_item_information");
    menuItemMinimize=
            Language.getCurrentLanguageInscription("menu_item_minimize");
    menuItemExit=Language.getCurrentLanguageInscription("menu_item_exit");
    menuItemOptions=Language.getCurrentLanguageInscription("menu_item_options");
    String[] menuItems={menuItemExit,menuItemMinimize,menuItemInformation,
        menuItemOptions,
    };
    menu=new List(Language.getCurrentLanguageInscription("menu_screen_title"),
            List.IMPLICIT,menuItems,null);
    menu.addCommand(commandShowMainCanvas);
    menu.setCommandListener(this);

    formInformation=new Form("PieceOfDictionary");
    formInformation.addCommand(commandShowMainCanvas);
    StringBuffer sbFormInformationText=new StringBuffer();
    sbFormInformationText.append(Language.getCurrentLanguageInscription(
            "help"));
    sbFormInformationText.append('\n');
    sbFormInformationText.append(Language.getCurrentLanguageInscription(
            "keys_remark"));
    sbFormInformationText.append('\n');
    Dictionary[] dictionaries=LexicalBase.getInstance().getDictionaries();
    if(dictionaries.length>1){
        StringBuffer sbDictionariesInformation=new StringBuffer();
        sbDictionariesInformation.append(dictionaries[0]);
        for(j=1;j<dictionaries.length;j++){
            sbDictionariesInformation.append(", ");
            sbDictionariesInformation.append(dictionaries[j]);
        }
        sbFormInformationText.append(replace(Language.
                getCurrentLanguageInscription("dictionaries_enumeration"),
                STRING_ESCAPE[0],sbDictionariesInformation.toString()));
    }else{
        sbFormInformationText.append(replace(Language.
                getCurrentLanguageInscription("dictionary_enumeration"),
                STRING_ESCAPE[0],dictionaries[0].toString()));
    }
    sbFormInformationText.append('\n');
    Language[] languages=Language.getLanguages();
    if(languages.length>1){
        StringBuffer sbLanguages=new StringBuffer();
        sbLanguages.append(languages[0]);
        for(j=1;j<languages.length;j++){
            sbLanguages.append(", ");
            sbLanguages.append(languages[j]);
        }
        sbFormInformationText.append(
                replace(Language.getCurrentLanguageInscription(
                "languages_enumeration"),STRING_ESCAPE[0],sbLanguages.toString()));
    }else{
        sbFormInformationText.append(
                replace(Language.getCurrentLanguageInscription(
                "language_enumeration"),STRING_ESCAPE[0],languages[0].toString()));
    }
    sbFormInformationText.append('\n');
    sbFormInformationText.append(replace(Language.getCurrentLanguageInscription(
            "money_request"),STRING_ESCAPE[0],MONEYBOXES));
    sbFormInformationText.append('\n');
    sbFormInformationText.append(getVersion(version));
    sbFormInformationText.append("\nquasiyoke@gmail.com");
    formInformation.append(sbFormInformationText.toString());
    formInformation.setCommandListener(this);

    formOptions=new Form(Language.getCurrentLanguageInscription(
            "options_screen_title"));
    formOptions.addCommand(commandShowMainCanvas);
    formOptions.addCommand(commandOptionsSave);
    formOptions.addCommand(commandSelectLanguage);
    formOptions.addCommand(commandOptionsSetDefaults);
    initializeOptionsFormItems();
    formOptions.setCommandListener(this);
}

/**
 * @return Является ли {@code character} пробельным символом в понимании
 * PieceOfDictionary.
 */
public static boolean isSpace(char character){
    return search(SPACE_CHARACTERS,character)>-1;
}

/**
 * Смешивает два цвета, учитывая их прозрачность.
 * 
 * @param coefficientFraction Количество {@code color1} в результате в
 * 2<sup>-{@link PieceOfDictionary#POINT_POSITION}</sup> долях.
 * Задаётся дробью: {@code (6&lt;&lt;PieceOfDictionary.POINT_POSITION)/7} --
 * 6/7; {@code (1&lt;&lt;PieceOfDictionary.POINT_POSITION)/10} -- 1/10.
 */
private static int middleColor(int color1,int color2,int coefficientFraction){
    coefficientFraction=coefficientFraction<=0?0:coefficientFraction>=1<<
            POINT_POSITION?1<<POINT_POSITION
            :coefficientFraction;
    int result=0;
    int j;
    for(j=0;j<32;j+=8){
        result|=(color1>>j&0xFF)*coefficientFraction+(color2>>j&0xFF)*((1<<
                POINT_POSITION)-coefficientFraction)>>POINT_POSITION<<j;
    }
    return result;
}

public static void minimize(){
    display.setCurrent(null);
}

public void pauseApp(){
}

/**
 * Печатает строку символов, не допуская выхода за пределы экрана (делая перенос
 * по словам).
 * 
 * @param POSITION Левый верхний угол надписи. Функция оставляет тут координаты
 * точки, в которой закончилась печать.
 */
public static void print(Graphics g,String string,final Point POSITION,Font font,
        int foregroundColor,Point screenSize){
    final int STRING_LENGTH=string.length();
    if(STRING_LENGTH==0){
        return;
    }
    final int FONT_HEIGHT=font.getHeight();
    final int LINE_RIGHT=screenSize.x-MARGIN_RIGHT;
    int j;
    int wordBeginIndex;
    int wordEndIndex=0;
    String word;
    int wordWidth;
    int wordLength;
    String subWord;
    boolean spacesHaveBegun;
    g.setFont(font);
    do{
        wordBeginIndex=wordEndIndex;
        spacesHaveBegun=false;
        while(++wordEndIndex<STRING_LENGTH){
            if(spacesHaveBegun){
                if(!isSpace(string.charAt(wordEndIndex))){
                    break;
                }
            }else{
                spacesHaveBegun=isSpace(string.charAt(wordEndIndex));
            }
        }
        word=string.substring(wordBeginIndex,wordEndIndex);
        while(true){
            wordWidth=font.stringWidth(word);
            if(POSITION.x+wordWidth>LINE_RIGHT){
                if(POSITION.x>MARGIN_LEFT){
                    POSITION.x=MARGIN_LEFT;
                    POSITION.y+=FONT_HEIGHT;
                /*
                 * Если слово находится в начале строки и не помещается, нужно
                 * разбить его на "подслова".
                 */
                }else{
                    wordLength=word.length();
                    j=wordLength;
                    do{
                        subWord=word.substring(0,--j);
                    }while(MARGIN_LEFT+font.stringWidth(subWord)>LINE_RIGHT);
                    g.setColor(COLOR_BACKGROUND);
                    g.fillRect(0,POSITION.y,screenSize.x,FONT_HEIGHT);
                    g.setColor(foregroundColor);
                    g.drawString(subWord,POSITION.x,POSITION.y,Graphics.LEFT|
                            Graphics.TOP);
                    POSITION.y+=FONT_HEIGHT;
                    word=word.substring(j,wordLength);
                }
            }else{
                if(POSITION.x==MARGIN_LEFT){
                    g.setColor(COLOR_BACKGROUND);
                    g.fillRect(0,POSITION.y,screenSize.x,FONT_HEIGHT);
                }
                g.setColor(foregroundColor);
                g.drawString(word,POSITION.x,POSITION.y,Graphics.LEFT|
                        Graphics.TOP);
                POSITION.x+=wordWidth;
                break;
            }
        }
    }while(wordEndIndex<STRING_LENGTH);
}

private static void quickSort(char[] array,int left,int right){
    char character;
    char temp;
    int j=left;
    int i=right;
    character=array[left+right>>1];
    while(j<=i){
        while(array[j]<character){
            j++;
        }
        while(character<array[i]){
            i--;
        }
        if(j<=i){
            temp=array[j];
            array[j]=array[i];
            array[i]=temp;
            j++;
            i--;
        }
    }
    if(left<i){
        quickSort(array,left,i);
    }
    if(j<right){
        quickSort(array,j,right);
    }
}

private static void quickSort(Comparable[] array,int left,int right){
    Comparable object;
    Comparable temp;
    int j=left;
    int i=right;
    object=array[left+right>>1];
    while(j<=i){
        while(array[j].compareTo(object)<0){
            j++;
        }
        while(object.compareTo(array[i])<0){
            i--;
        }
        if(j<=i){
            temp=array[j];
            array[j]=array[i];
            array[i]=temp;
            j++;
            i--;
        }
    }
    if(left<i){
        quickSort(array,left,i);
    }
    if(j<right){
        quickSort(array,j,right);
    }
}

/**
 * Ищет элемент в отсортированном массиве.
 * 
 * @return Индекс вхождения в массиве, или -1 в случае отсутствия элемента там.
 */
public static int search(char[] array,char character){
    if(array.length==0){
        return -1;
    }
    int middle;
    int begin=0;
    int end=array.length-1;
    do{
        middle=begin+end>>1;
        if(character==array[middle]){
            return middle;
        }
        if(character<array[middle]){
            end=middle-1;
        }else{
            begin=middle+1;
        }
    }while(begin<=end);
    return -1;
}

private void readParameters(){
    int j;
    XmlPullParser parser=null;
    try{
        parser=xmlPullParserFactory.newPullParser();
        parser.setInput(
                this.getClass().getResourceAsStream(PARAMETERS_FILE_PATH),null);
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- parser initialization: "+e.
                toString());
    }
    try{
        parser.nextTag();
        parser.require(XmlPullParser.START_TAG,"","PieceOfDictionary_parameters");
        boolean necessaryAttributeFound=false;
        int attributesCount=parser.getAttributeCount();
        for(j=0;j<attributesCount;j++){
            if("version".equals(parser.getAttributeName(j))){
                int[] parametersVersion=getVersion(parser.getAttributeValue(j));
                if(compareVersions(parametersVersion,version)==1){
                    showError(
                            "Parameters version (v. "+
                            parser.getAttributeValue(j)+
                            ") is fresher than PieceOfDictionary's one (v. "+
                            getVersion(version)+
                            "). PieceOfDictionary will try to read this parameters -- errors are possible.");
                }
                necessaryAttributeFound=true;
                break;
            }
        }
        if(!necessaryAttributeFound){
            showError("Parameters reading -- file version not found.");
        }
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- reading file's head -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showErrorAndExit("Parameters reading -- reading file's head: "+
                e.toString());
    }

    try{
        parser.nextTag();
        parser.require(XmlPullParser.START_TAG,"","dictionaries");
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- reading dictionaries information head -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showErrorAndExit("Parameters reading -- reading dictionaries information head: "+
                e.toString());
    }
    Vector dictionariesPaths=new Vector();
    try{
        while(parser.nextTag()==XmlPullParser.START_TAG){
            parser.require(XmlPullParser.START_TAG,"","dictionary_path");
            dictionariesPaths.addElement(parser.nextText());
            parser.require(XmlPullParser.END_TAG,"","dictionary_path");
        }
    }catch(XmlPullParserException e){
        showError("Parameters reading -- reading dictionaries information -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showError("Parameters reading -- reading dictionaries information: "+e.
                toString());
    }
    if(dictionariesPaths.isEmpty()){
        showErrorAndExit("Parameters reading: no dictionaries found.");
    }
    try{
        parser.require(XmlPullParser.END_TAG,"","dictionaries");
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- reading dictionaries information tail -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showErrorAndExit("Parameters reading -- reading dictionaries information tail: "+
                e.toString());
    }

    try{
        parser.nextTag();
        parser.require(XmlPullParser.START_TAG,"","languages");
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- reading languages information head -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showErrorAndExit("Parameters reading -- reading languages information head: "+
                e.toString());
    }
    Vector languagesPaths=new Vector();
    try{
        while(parser.nextTag()==XmlPullParser.START_TAG){
            parser.require(XmlPullParser.START_TAG,"","language_path");
            languagesPaths.addElement(parser.nextText());
            parser.require(XmlPullParser.END_TAG,"","language_path");
        }
    }catch(XmlPullParserException e){
        showError("Parameters reading -- reading languages information -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showError("Parameters reading -- reading languages information: "+e.
                toString());
    }
    if(languagesPaths.isEmpty()){
        showErrorAndExit("Parameters reading: no languages found.");
    }

    LexicalBase.setDictionariesPaths(dictionariesPaths);
    Language.setLanguagesPaths(languagesPaths);

    try{
        parser.require(XmlPullParser.END_TAG,"","languages");
    }catch(XmlPullParserException e){
        showErrorAndExit("Parameters reading -- reading languages information tail -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showErrorAndExit("Parameters reading -- reading languages information tail: "+
                e.toString());
    }
    try{
        parser.nextTag();
        parser.require(XmlPullParser.END_TAG,"","PieceOfDictionary_parameters");
    }catch(XmlPullParserException e){
        showError("Parameters reading -- file's tail -- file structure problem: "+
                e.toString());
    }catch(IOException e){
        showError("Parameters reading -- file's tail: "+e.toString());
    }
}

/**
 * @throws RecordStoreException Если попытка чтения хранилища завершилась
 * неудачей.
 */
private void readStorage() throws RecordStoreException{
    int j;
    int i;
    RecordStore recordStore=null;
    byte[] record;
    boolean resavingNeeds=false;
    try{
        recordStore=RecordStore.openRecordStore(RECORD_STORE_NAME,false);
    }catch(RecordStoreException e){
        throw new RecordStoreException(RECORD_STORE_NOT_FOUND_MESSAGE);
    }
    try{
        record=recordStore.getRecord(1);
    }catch(RecordStoreException e){
        try{
            recordStore.closeRecordStore();
        }catch(RecordStoreException f){
        }
        throw e;
    }
    try{
        recordStore.closeRecordStore();
    }catch(RecordStoreException e){
        throw e;
    }
    DataInputStream dataInputStream=new DataInputStream(
            new ByteArrayInputStream(record));
    int[] storageVersion=new int[version.length];
    try{
        for(j=0;j<storageVersion.length;j++){
            storageVersion[j]=dataInputStream.readInt();
        }
    }catch(IOException e){
        throw new RecordStoreException("reading storage version: "+e.toString());
    }
    if(compareVersions(storageVersion,version)==1){
        showError(
                "Store keeps records of PieceOfDictionary "+getVersion(
                storageVersion)+" -- fresher than this ("+getVersion(version)+
                "). PieceOfDictionary will try to read this records -- errors are possible.");
        resavingNeeds=true;
    }
    char[][] keysChars=new char[KEYS_COUNT][];
    int keyCharsCount;
    try{
        for(j=0;j<KEYS_COUNT;j++){
            keyCharsCount=dataInputStream.readByte();
            if(keyCharsCount>MAX_KEY_CHARS_COUNT){
                throw new RecordStoreException();
            }
            keysChars[j]=new char[keyCharsCount];
            for(i=0;i<keyCharsCount;i++){
                keysChars[j][i]=dataInputStream.readChar();
            }
        }
    }catch(IOException e){
        throw new RecordStoreException("reading keys chars: "+e.toString());
    }
    String languageName;
    try{
        languageName=dataInputStream.readUTF();
    }catch(IOException e){
        throw new RecordStoreException("reading language name: "+e.toString());
    }
    boolean mainCanvasFullscreen;
    try{
        mainCanvasFullscreen=dataInputStream.readBoolean();
    }catch(IOException e){
        throw new RecordStoreException("reading screen state: "+e.toString());
    }
    /*
     * Присваивания полям выполнять ПОСЛЕ всех операций, могущих
     * выбросить исключение.
     */
    this.keysChars=keysChars;
    MainCanvas.fullscreen=mainCanvasFullscreen;
    Language.setCurrent(languageName);
    if(resavingNeeds){
        deleteStorageAndSaveOptions();
    }
}

public static String replace(String string,String oldPiece,String newPiece){
    int oldPieceBeginIndex=string.indexOf(oldPiece,0);
    StringBuffer sbString=new StringBuffer();
    if(oldPieceBeginIndex==-1){
        sbString.append(string);
        sbString.append(newPiece);
    }else{
        sbString.append(string.substring(0,oldPieceBeginIndex));
        sbString.append(newPiece);
        sbString.append(string.substring(oldPieceBeginIndex+oldPiece.length()));
    }
    return sbString.toString();
}

public void showMenu(){
    display.setCurrent(menu);
}

public static void showError(String message){
    Alert alert=new Alert("PieceOfDictionary",message,null,AlertType.ERROR);
    Displayable currentDisplayable=display.getCurrent();
    alert.setTimeout(ALERT_TIMEOUT);
    if(currentDisplayable==null||currentDisplayable instanceof Alert){
        display.setCurrent(alert);
        try{
            Thread.sleep(ALERT_TIMEOUT);
        }catch(InterruptedException e){
        }
    }else{
        display.setCurrent(alert,currentDisplayable);
    }
    System.out.println(message);
}

public static void showErrorAndExit(String message){
    Alert alert=new Alert("PieceOfDictionary",message,null,AlertType.ERROR);
    alert.setTimeout(ALERT_TIMEOUT);
    display.setCurrent(alert);
    try{
        Thread.sleep(ALERT_TIMEOUT);
    }catch(InterruptedException e){
    }
    System.out.println(message);
    instance.destroyApp(false);
}

private void showLanguageSelect(boolean canShowMainCanvas){
    if(languageSelect==null){
        languagesNames=Language.getLanguagesNames();
        languageSelect=new List("Select language:",List.IMPLICIT,languagesNames,
                null);
        languageSelect.addCommand(commandExit);
        languageSelect.setCommandListener(this);
    }
    if(canShowMainCanvas){
        languageSelect.addCommand(commandShowMainCanvas);
    }else{
        languageSelect.removeCommand(commandShowMainCanvas);
    }
    display.setCurrent(languageSelect);
}

private void showMainCanvas(){
    display.setCurrent(MainCanvas.getInstance());
}

/**
 * Сортирует массив в возрастающем порядке.
 */
public static char[] sort(char[] array){
    if(array.length>1){
        quickSort(array,0,array.length-1);
    }
    return array;
}

/**
 * Сортирует массив в возрастающем порядке.
 */
public static Comparable[] sort(Comparable[] array){
    if(array.length>1){
        quickSort(array,0,array.length-1);
    }
    return array;
}

/**
 * Читает из RMS (а при первом запуске записывает в RMS) параметры.
 * <p>
 * RMS_запись ::= Major_версия Minor_версия Номер_ревизии (Число_букв_клавиши
 * (Буква)(Число_букв_клавиши))({@link PieceOfDictionary#KEYS_COUNT})
 * <p>
 * Major_версия -- 4 байт ({@code int}).<br>
 * Minor_версия -- 4 байт ({@code int}).<br>
 * Номер_ревизии -- 4 байт ({@code int}).<br>
 * Число_букв_клавиши -- 1 байт. Количество букв за этой клавишей.<br>
 * Буква -- 2 байт ({@code char}).<br>
 */
public void startApp(){
    /*
     * Инициализацию много чего лучше делать тут: из конструктора плохо
     * вызывается notifyDestroyed(), например.
     */
    if(instance==null){
        version=getVersion(getAppProperty("MIDlet-Version"));
        display=Display.getDisplay(this);
        instance=this;
        readParameters();
        try{
            readStorage();
            LexicalBase.getInstance().matchKeysCharsAndAlphabets(keysChars);
            showMainCanvas();
        }catch(RecordStoreException e){
            if(!e.getMessage().equals(RECORD_STORE_NOT_FOUND_MESSAGE)){
                showError(
                        "Store reading -- \""+e.toString()+
                        "\". PieceOfDictionary will try to rewrite storage with defaults.");
            }
            deleteStorageAndSaveDefaults();
        }
    }else{
        MainCanvas.getInstance().checkSize();
        showMainCanvas();
    }
}

}
